!
!  Module      : omatt_mod
!
! Authors     : Tim Harries
!             : Andrew King (MSSL)
!          
!
! Description : Convert from OM detector to sky coordinates.
! 
! Comments    : 
!
! $Id: module.f90,v 1.4 1997/12/18 16:01:10 gvacanti Exp $
! 
! History : Created  : March 98 - TJH
!           Modified : 21 May 98 - Added use sla.
!           Modified : 27 May 99 - Updated parameter names.
!           Modified : 1st June 00 - Added calculation of galactic coordinates for RA and DEC
!           Modified : 1st June 00 - Added addition of two extra keywords and columns to hold
!                                    the galactic coordinates.
!           Modified : 2nd June 00 - Changed the galactic coordinates to 8 byte data
!           Modified : 2nd June 00 - Installed code to release memory that had been allocated
!                                    this was left hanging allocated.
!           Modified : 4th june 00 - Changed RA and DEC to 8 byte data
!
!
!           Modified : 21/06/2000 Changed poserr to be consistent with arcsecs definition from OMDETECT.
!
!           Modified : 1/8/2000  (CB) - Added -90 degree image rotation
!                      to correct for error in coordinate system.
!           Modified : 14/8/2000 (CB) - Maths functions sin and cos replaced by dsin and 
!                                       dcos (double precision versions),
!                                       & some other changes to double precision made.
!           Modified : 15/8/2000 (CB) - Output image corrected for image distortion.
!           Modified : 21/8/2000 (CB) - Function "MissionTime" added to compute the difference
!                                       between the average of the start and end observations
!                                       currently fixed at 0.4765
!          Modified : 24/8/2000 (CB) -  Some write statements deleted. The +ve signs for 
!                                       the distortion corrections changed to -ve.
!          Modified : 07/09/2000     -  Attitude matrinningix read in. Boresight calculation changed -
!                                       uses [A*(B)^-1]*(1,0,0).
!          Modified : 28/09/2000(AMK)-  Changed parameters in call message!
!          modified : 02/10/2000(AMK)-  Reinstated the call to CAL_omGetPLateScalePnte Rotate
!
!          modified : 02/10/2000(SRR)-  Inserted inv_boremat for the inverted boresight matrix
!
!          modified : 05/10/2000(AMK)-  Replaced the call getFIlterID in mssllib by the OAL call filterIdToInt
!          modified : 10/10/2000(CB) -  Paramter "usecat" set to false if it is true- not yet implemented
!          modified : 20/12/2000(CB) -  Removed world-coordinate attributes from input image header file.
!          modified : 18/01/2000(CB) -  Exposure array added to the PPS output file. Code for astrometry
!                                       correction moved to a subroutine. 
!          modified : 24/05/2001(CB) -  Attributes "RA_SCX, DEC_SCX & PA added to output file- values
!                                       refer to the spacecraft pointing position. RA_PNT & RA_DEC
!                                       also added.  
!          modified : 14/06/2001(CB) -  Removed code that changes position errors (SPR 2429)
!          modified : 02/08/2001(CB) -  Removed definition of struct yySpaceCraftAttitudeType-
!                                       defined in caloaldefs.  Replaced the matrix attmatrix in the 
!                                       call to OAL_toAttitudeMatrix by attmatrix1 (correct subscripts)
!                                       Changed program so that all 1-d arrays start at 1. Tidied
!                                       program up- removed astrometry correction code into a
!                                       separate module- added module "utility" which contains
!                                       some subroutines for doing routine jobs. 
! 
!          modified : 10.09.2001(VNY) - The centre of the CCD is fixed at (1024.5, 1024.5). Output RA and 
!                                       DEC coordinates has been recalculated according to the origin of the
!                                       CCD coordinate system lying at the center of the first pixel.
!                                       Correspondingly, the output image shift and rotation have
!                                       been corrected                            
!          modified : 16.11.2001 (CB) - Code to store an exposure-image map removed. Rotated image pixels
!                                       outside the image boundary set to NULL
!          modified : 16/12/2001(CB) -  Parameter "usecat" reactivated and astrometry correction now done
!                                       if sufficient stars   
!          modified : 17/09/2002 (VNY) - adjusted to handle with the source list without the MAG-column 
!                                       (omdetect now does not generate it). If there is no MAG 
!                                       column in the SRCLIST-file, the magnitudes are set to zero 
!
!         modified : 25/09/2002 (CB) - Position angles of each source adjusted for image rotation 
!         modified : 05/11/2002 (CB) - pgplot code removed
!         modified : 02/12/2002 (CB) - added boolean parameter to control production of the 
!                                      rotated image - (false no image- defaults to true)   
!*****************************************************************************************  
module omatt_mod

  use param
  use oal
  use cal
  use dal
  use sasdal
  use caloaldefs

  use errorHandling
  use sla
  use readModes_mod
  use omallocation_mod 


  use omconstants_mod
  use omastromcorrect_mod    ! For doing the astrometry correction
 
  use integration_mod
  use omdistortion_mod 
  use omrotation_mod
  use omcelestial_mod
implicit none


public
   public omatt

contains

!*******************************************************************
! subroutine omatt
! NB Paramters xCentre & yCentre are stored in module omconstants_mod
! in mssllib and are 1024.5 * 1024.5
! NB the input pixel coordinates are defined such that the centre
! of the bottom-left pixel is 1, 1
!******************************************************************
subroutine omatt


    type(DataSetT) :: inDataSet                                       ! Input dataset handle.
    type(DataSetT) :: sourceListDataSetH                              ! Source List dataset handle.
  
    type(TableT) :: sourceListTableH                                  ! Source List table handle.

    character(255) :: sourceListSet                               ! OM OSW Source List filename
    character(10)           :: datamode   ! datamode attribute
    character(6)            :: omFilter   ! the OM filter
    integer                 :: xBinning
    integer                 :: yBinning
    integer             :: nrows              ! the number of rows in the table
    integer             :: xbin, ybin         ! binning factors
    integer             :: xsize, ysize       ! size of the image
    integer                     :: OM_MODE    ! the om mode
    real(double)  :: rotationCorrection = 90.0_double! Correction for coordinate error
    real(double)  :: plateScale       ! the platescale
    real(double) :: ra_point = 0.0_double, dec_point= 0.0_double        ! position of the OM tangent point

               
    real(single), dimension(:), pointer  ::  xpos => null()
    real(single), dimension(:), pointer  ::  ypos => null()

    real(double) :: ra_scx, dec_scx, pa_scx  ! The ra, dec and pa of the spacecraft     
    real(double), dimension(:), pointer :: rightAscension => null()
    real(double), dimension(:), pointer :: declination => null()
    real(single), dimension(:), pointer :: posErr => null()
    real(double), pointer :: galacticLong(:) => null()  ! Array for galactic longitudes  
    real(double), pointer :: galacticLat(:)  => null()  ! Array for galactic latitudes
  
    real(single), dimension(:), pointer :: magnitude => null()
    real(double) :: rollangle                  ! the roll angle
                        
  
    real(double)  :: lambda           ! wavelength for CAL
    real(single)  :: nullValue
    logical(Bool) :: binBPE                    ! BPE binning on?
    logical(bool) :: useCat                        ! Use the USNO-SA 1 catalog.
    
    integer                  :: istat              ! status for memory allocation
    integer(int32)           :: filterIdInt
    
    integer(int32) :: position
 
    character(255) :: set                                         ! OM image intermediate filename
    integer(int32) :: i, x01, y01



    nullValue = real(-99999)


    !*************************************
    ! Get the command-line parameters
    !*************************************
    set = stringParameter("set")
    inDataSet = DataSet(set, read) 


    !***********************************************************
    ! Read in WINDOWXO (x0) & WINDOWY) (y0) and the image size
    !***********************************************************
    call readModes(inDataSet, x01, y01, xsize, ysize)


    x0 = x01
    y0 = y01   


    sourceListSet = stringParameter("sourcelistset")
    useCat = booleanParameter("usecat")


    !**************************
    ! Access the source list
    !**************************
    sourceListDataSetH = dataSet(sourceListSet, modify)      ! get osw list handle

   
    xbin = int32Attribute(sourceListDataSetH, "BINAX1")      ! binning factors
    ybin = int32Attribute(sourceListDataSetH, "BINAX2")
 
    if (xbin .ne. ybin) then
        call fatal("diffBin", "This version cannot deal with different binning factors in x and y")
    end if

    binBPE = booleanAttribute(sourceListDataSetH, "BINBPE")
 
    ! Calculate the image binning.
    xBinning = 2 ** xbin
    yBinning = 2 ** ybin


    if (binBPE) then
        xBinning = xBinning * 2
        yBinning = yBinning * 2
    end if


    dXbinning = dble(xBinning)  ! Need the double values of xBinning and 
    dYbinning = dble(yBinning)  ! yBinning fairly frequently
   
    call InitialiseDistortion(x01, y01, xBinning, yBinning)

    omFilter = stringAttribute(sourceListDataSetH, "FILTER")
 
    dataMode = stringAttribute(sourceListDataSetH, "DATAMODE")  ! needed for cal

    if (trim(dataMode) .eq. "IMAGING") then 
       OM_MODE = IMAGE
    else
       OM_MODE = FAST
    end if
   
    sourceListTableH = table(sourceListDataSetH, "SRCLIST")     ! get the source list

    nrows = numberOfRows(sourceListTableH)

    !*********************************************************
    ! Get the x and y pixel coordinates of each source
    ! NB these are stored as real32 values.
    !********************************************************

    xPos => allocateSingle(1, nRows, "xpos")
 
    call ReadReal32Column(sourceListTableH, "XPOS", nRows, 0, xpos)

    yPos => allocateSingle(1, nRows, "ypos")
  
    call ReadReal32Column(sourceListTableH, "YPOS", nRows, 0, ypos)
  


    magnitude => AllocateSingle(1, nRows, "magnitude")
    poserr => allocateSingle(1, nRows, "poserr")
   

    !----------------- VNY 17.09.2002 Checking if the MAG column does exist 
    if (hasColumn(sourceListTableH, "MAG")) then
        call ReadReal32Column(sourceListTableH, "MAG", nRows, 0, posErr)
        magnitude = real(posErr)      
    else 
       magnitude = 0.0
    endif
    !----------------------
 
  
    call ReadReal32Column(sourceListTableH, "POSERR", nRows, 0, posErr)
 
 

    !*******************************************************
    ! get the astrometry info keywords from the source list
    !*******************************************************
    ra_scx = real64Attribute(sourceListDataSetH, "RA_SCX")   ! Spacecraft RA in radians
    dec_scx = real64Attribute(sourceListDataSetH, "DEC_SCX") ! Spacecraft dec in radians

    pa_scx = real64Attribute(sourceListDataSetH, "PA")       ! Spacecraft PA in degrees
   



    ra_scx = ra_scx * degreesToRadians
    dec_scx = dec_scx * degreesToRadians

 
    !********************************************
    ! Transform the source position angles
    ! to the equatorial coordinate system
    !*******************************************

    call ModifySourcePositionAngles(sourceListTableH, pa_scx + rotationCorrection, nullValue)

    pa_scx = pa_scx * degreesToRadians
    rollAngle = pa_scx


    !**********************************************
    ! rotationCorrection corrects for OM axes definition.
    !**********************************************
    rotationCorrection = rotationCorrection * degreesToRadians 
    rollAngle = rollAngle + rotationCorrection
   


   
    !************************************
    ! Get the filter Id integer value
    !************************************
    filterIdInt = filterIdToInt(omFilter)

    !*********************
    ! Set the CAL state
    !*********************
    call CAL_setState(OM, ccdModeId=OM_MODE, filterId = filterIdInt)


    !******************************************
    ! Boresight stuff
    !******************************************


    call OmPointingDirection(inDataSet, ra_point, dec_point)

    call release(inDataSet)


 

    call CAL_omGetPlateScale(lambda, plateScale)  ! Get the plate-scale (arcsecs per pixel) 
    plateScale = plateScale * degreesToRadians / 3600.0_double         ! platescale converted to radians



  
    sinRollAngle = dsin(rollAngle)  ! Need these quantities quite a lot
    cosRollAngle = dcos(rollAngle)


    rightAscension => AllocateDouble(1, nRows, "rightAscension")
    declination => AllocateDouble(1, nRows, "declination")
    galacticLong => AllocateDouble(1, nrows, "galacticLong")
    galacticLat => AllocateDouble(1, nrows, "galacticLat")
 

    !*********************************************************************
    ! For each source compute its undistorted position 
    ! relative to the centre of the full-frame image, and then
    ! its RA, dec, galactic longitude and latitude
    !*********************************************************************


    do i = 1, nrows

         call ComputeCelestialCoordinates(xPos(i), yPos(i), x0, y0,   & 
                                       real(xBinning), real(yBinning), &
                                       plateScale, rollAngle, &
                                       ra_point, dec_point, &
                                       rightAscension(i), declination(i) , &
                                       galacticLong(i), galacticLat(i)) 

    end do


    !*********************************************************
    ! Add RA, dec, galactic longitude and galactic latitude 
    ! columns to the fits source-list file
    !*********************************************************
    position = columnNumber(sourceListTableH, "POSERR") + 1

    call AddReal64Column(sourceListTableH, rightAscension * radiansToDegrees, &
             nrows, "RA", "degrees", "Not corrected for boresight offset", nullvalue, position)
    call AddReal64Column(sourceListTableH, declination * radiansToDegrees, &
             nrows, "DEC", "degrees", "Not corrected for boresight offset", nullValue, position + 1)
    call AddReal64Column(sourceListTableH, galacticLong  * radiansToDegrees, &
                   nrows, "LII", "degrees", "Not corrected for boresight offset", nullValue, position + 2)
    call AddReal64Column(sourceListTableH, galacticLat  * radiansToDegrees, &
                    nrows, "BII", "degrees", "Not corrected for boresight offset", nullValue, position + 3)



    !****************************************************************
    ! We only do the next bit if we have access to the usno catalogue
    !****************************************************************

    if (useCat) then
        ! NB all angles should be in radians, apart from poserr
        ! which is in arc secs
   
        call CorrectTheAstrometry(nrows,  rightAscension, declination, &
            posErr, magnitude, ra_point, dec_point, sourceListSet)     

       
    end if
   


    deallocate(rightAscension, declination, posErr, &
                magnitude, galacticLong, galacticLat, stat = istat)
    call CheckDeallocation(istat, "rightAscension")

 
    call SetFitsTableAttributes(sourceListDataSetH, ra_point, dec_point)


    call release(sourceListTableH)
    call release(sourceListDataSetH)

 


    if (booleanParameter("rotateimage")) then  ! Create the rotated image
        ! Rotate the image by rollAngle (radians)
        call CreateRotatedImage(trim(stringParameter("set")), &
                                trim(stringParameter("ppsoswset")), &
              rollAngle, ra_point, dec_point, &
              xBinning, yBinning, ra_scx, dec_scx, plateScale)
    end if

  
    !**************************************
    ! Release memory allocation
    !**************************************
    


    deallocate(rawx, rawy, xpos, ypos, stat = istat)
    call CheckDeallocation(istat, "xpos")

 
end subroutine omatt

!*******************************************
! Function MissionTime - Returns the number
! of seconds from the start of the mission
! to the observation time. Added by CB because
! the call to CAL_GetBoresiteMatrix needs
! the time in seconds from the launch.
!******************************************
real(double) function  MissionTime(primaryTable)
    type(DataSetT), intent(in)   :: primaryTable

    character(19)  :: dateObs         

    real(double)  :: JulianDate0, JulianDate1 ! , JulianDate2
    character(19)  :: missionReferenceTime = "1998-01-01T00:00:00"

    ! Compute the Julian date of the mission reference time.
    JulianDate0 = ModifiedJulianDate(missionReferenceTime)

    ! Compute the Julian date of the start of the observations.
    dateObs = stringAttribute(primaryTable,  "DATE-OBS")    
    JulianDate1 = ModifiedJulianDate(dateObs)


   MissionTime = (JulianDate1 - JulianDate0) * 86400.0d0


end function MissionTime


!********************************************
! Subroutine ModifySourcePositionAngles
!*******************************************
subroutine ModifySourcePositionAngles(tableSetH, angle, nullValue)
    type(TableT), intent(in) :: tableSetH
    real(double), intent(in) :: angle
    real(single), intent(in) :: nullValue
    
    real(single), pointer :: positionAngle(:) => null()
    integer(int32) :: i, noRows, position
  
    noRows = numberOfRows(tableSetH)

    position = columnNumber(tableSetH, "PA")
 
    positionAngle => allocateSingle(1, noRows, "positionAngle")
  
    call ReadReal32Column(tableSetH, "PA", noRows, 0, positionAngle)
 

 
    call deleteColumn(tableSetH, "PA")


    do i = 1, noRows
        if (.not. isNullValue(positionAngle(i))) then 
            positionAngle(i) = positionAngle(i) + angle
            if (positionAngle(i) .gt. 360.0) positionAngle(i) = positionAngle(i) - 360.0
            if (positionAngle(i) .gt. 180.0) positionAngle(i) = positionAngle(i) - 180.0
        end if
    end do
    call AddReal32Column(tableSetH, positionAngle, noRows, "PA", "degrees", &
             "PA of major-axis Referred to equatorial system (measured anti-clockwise)", nullvalue, position)
 

    deallocate(positionAngle)
end subroutine ModifySourcePositionAngles

!***************************************************************
!
!**************************************************************
subroutine OmPointing(dataSet)
    type(dataSetT), intent(in) :: dataSet

    real(double) :: ra_point, dec_point        ! position of the OM tangent point
    real(double) :: rollangle                  ! the roll angle
    real(double) :: time                        
    real(double) :: attmatrix(1 : 3, 1 : 3)    ! rotation matrix
    real(double) :: attmatrix1(0 : 2, 0 : 2)    ! rotation matrix

    real(double) :: att_invbor(3, 3)           !matrix for product of attitude*boresight
    real(double) :: ompoint(3)=(/1.0_double, 0.0_double, 0.0_double/)         !om boresight vector in OM frame
    real(double) :: pointvec(3)        ! s/c pointing
    real(double) :: pointvec2(3)       
    real(double), pointer :: boremat(:,:)       ! boresight offset matrix
    real(double) :: inv_boremat(3, 3)          ! matrix for inverse of boremat
    real(double) :: omvec(3)                   ! om pointing vector
  

    real(double) :: ra_scx, dec_scx, pa_scx  ! The ra, dec and pa of the spacecraft     
    real(double) :: r1                         ! modulus
    type (SpacecraftAttitudeType) attitude
     
    
    integer i, j    

    !***********************************************************
    ! Get the number of seconds since the mission reference time
    !***********************************************************
    time =  MissionTime(dataSet)


    !*******************************************************
    ! get the astrometry info keywords from the source list
    !*******************************************************
    ra_scx = real64Attribute(dataSet, "RA_SCX") * degreesToRadians   ! Spacecraft RA in radians
    dec_scx = real64Attribute(dataSet, "DEC_SCX") * degreesToRadians ! Spacecraft dec in radians
    pa_scx = real64Attribute(dataSet, "PA")        ! Spacecraft PA in degrees
 
 
    pa_scx = pa_scx * degreesToRadians
    rollAngle = pa_scx

    
    call CAL_GetBoresightMatrix(time, boremat) ! boresight offset matrix
    
    !*******************************************************************
    ! need inverted boresight matrix which is obtained via the transpose. 
    !********************************************************************
    inv_boremat = TRANSPOSE(boremat)

    attitude = SpaceCraftAttitudeType(ra_scx, dec_scx, rollangle) 

    !*******************************************************
    ! This routine requires a matrix of size 0:2, 0:2, but
    ! after that we want it to be of size 1:3, 1:3
    !******************************************************
    call OAL_toAttitudeMatrix(attitude, attmatrix1)
    
    do i = 1,  3 
        do j = 1, 3 
            attmatrix(i, j) = attmatrix1(i - 1, j - 1)
        end do
    end do

    !******************************************************************************
    ! converts RA and DEC from spherical cordinates to cartesian coordinates vector
    !*****************************************************************************
    call sla_dcs2c(ra_scx, dec_scx, pointvec)

    !********************************************************
    ! normalize the vector pointvec and return in pointvec2 -
    ! r1 holds the mdulus of the vector. pointvec2 will store
    ! the unit-vector in the direction of pointvec
    !********************************************************
    call sla_dvn(pointvec, pointvec2, r1)

    !
    !
    ! use SIMON'S BIT
    !
    !write(*,*)'using SIMONS BIT (essentially redundant now but keep for now
    !--------------------------------------------------------------------------------------------------------
    pointvec(1) = -pointvec2(1)
    pointvec(2) = -pointvec2(2)
    pointvec(3) =  pointvec2(3)
    !--------------------------------------------------------------------------------------------------------

 

    ! convert euler axis vector to a rotation matrix
     

    ! initialise output vector
    omvec = 0.0_double


!**********************************************************************************************
!  To get boresight correction, apply attitudematrix*(boresightmatrix)^(-1)*ompoint(=1,0,0)
!  Boresight * vector in SC frame gives vector in OM frame so boresight^(-1)*vector in OM frame gives 
!  vector in SC frame.  Then attitude * vector in S/C frame gives vector in Equatorial Reference Frame.
!
!  This creates omvec, the om boresight vector measured in the ERF
!  Note this process is applied here by first doing attmatrix*inv_boremat, then multiplying the 
!  ompoint vector by the product and decomposing to RA and DEC. The ompoint vector 
!  is the OM boresight as measured in the OM reference frame
!*********************************************************************************************

    att_invbor = matmul(attmatrix, inv_boremat)
    omvec = matmul(att_invbor, ompoint)

    !************************************************************************
    ! converts the omvec cartesian vector to ra and dec spherical coordinates
    ! and check RA lies within sensible range
    ! ra_point and ra_dec refer to the pointing direction of the instrument
    !************************************************************************

    call sla_dcc2s(omvec, ra_point, dec_point)

    if (ra_point .lt. 0.0_double) ra_point = ra_point + 2.0_double * pi
   
  
    write(errstr, '(a, f8.4)') "RA of OM boresight: ", ra_point * radiansToDegrees
    call message(AppMsg, Verbose)
    write(errstr, '(a, f8.4)') "DEC of OM boresight: ", dec_point * radiansToDegrees
    call message(AppMsg, Verbose)


end subroutine OmPointing




subroutine OmPointing1(dataSet)
    type(dataSetT), intent(in) :: dataSet

    real(double) :: ra_point, dec_point        ! position of the OM tangent point
    real(double) :: rollangle                  ! the roll angle
   
    real(double) :: attmatrix(1 : 3, 1 : 3)    ! rotation matrix
    real(double) :: attmatrix1(0 : 2, 0 : 2)    ! rotation matrix

    real(double) :: att_invbor(3, 3)           !matrix for product of attitude*boresight
    real(double) :: ompoint(3)=(/1.0_double, 0.0_double, 0.0_double/)         !om boresight vector in OM frame
    real(double) :: pointvec(3)        ! s/c pointing
    real(double) :: pointvec2(3)       
    real(double), pointer :: boremat(:,:)       ! boresight offset matrix
    real(double) :: inv_boremat(3, 3)          ! matrix for inverse of boremat
    real(double) :: omvec(3)                   ! om pointing vector

    real(double) :: ra_scx, dec_scx, pa_scx  ! The ra, dec and pa of the spacecraft     
    real(double) :: r1                         ! modulus
    type (SpacecraftAttitudeType) attitude
     
    integer i, j    
    real(double) :: time


    !*******************************************************
    ! get the astrometry info keywords from the source list
    !*******************************************************
    ra_scx = real64Attribute(dataSet, "RA_SCX") * degreesToRadians   ! Spacecraft RA in radians
    dec_scx = real64Attribute(dataSet, "DEC_SCX") * degreesToRadians ! Spacecraft dec in radians
    pa_scx = real64Attribute(dataSet, "PA")        ! Spacecraft PA in degrees
 
 
    pa_scx = pa_scx * degreesToRadians
    rollAngle = pa_scx



    !***********************************************************
    ! Get the number of seconds since the mission reference time
    !***********************************************************
    time =  MissionTime(dataSet)

   

    call CAL_GetBoresightMatrix(time, boremat) ! boresight offset matrix
   
    
    !*******************************************************************
    ! need inverted boresight matrix which is obtained via the transpose. 
    !********************************************************************
    inv_boremat = TRANSPOSE(boremat)

    attitude = SpaceCraftAttitudeType(ra_scx, dec_scx, rollangle)

    !*******************************************************
    ! This routine requires a matrix of size 0:2, 0:2, but
    ! after that we want it to be of size 1:3, 1:3
    !******************************************************
    call OAL_toAttitudeMatrix(attitude, attmatrix1)
    
    forall( i = 1 : 3) 
        forall(j = 1 : 3) attmatrix(i, j) = attmatrix1(i - 1, j - 1)
    end forall

   
 
    !******************************************************************************
    ! converts RA and DEC from spherical cordinates to cartesian coordinates vector
    !*****************************************************************************
    call sla_dcs2c(ra_scx, dec_scx, pointvec)

    !********************************************************
    ! normalize the vector pointvec and return in pointvec2 -
    ! r1 holds the mdulus of the vector. pointvec2 will store
    ! the unit-vector in the direction of pointvec
    !********************************************************
    call sla_dvn(pointvec, pointvec2, r1)

    !
    !
    ! use SIMON'S BIT
    !
    !write(*,*)'using SIMONS BIT (essentially redundant now but keep for now
    !--------------------------------------------------------------------------------------------------------
    pointvec(1) = -pointvec2(1)
    pointvec(2) = -pointvec2(2)
    pointvec(3) =  pointvec2(3)
    !--------------------------------------------------------------------------------------------------------

 

    ! convert euler axis vector to a rotation matrix
     

    ! initialise output vector
    omvec = 0.0_double


!**********************************************************************************************
!  To get boresight correction, apply attitudematrix*(boresightmatrix)^(-1)*ompoint(=1,0,0)
!  Boresight * vector in SC frame gives vector in OM frame so boresight^(-1)*vector in OM frame gives 
!  vector in SC frame.  Then attitude * vector in S/C frame gives vector in Equatorial Reference Frame.
!
!  This creates omvec, the om boresight vector measured in the ERF
!  Note this process is applied here by first doing attmatrix*inv_boremat, then multiplying the 
!  ompoint vector by the product and decomposing to RA and DEC. The ompoint vector 
!  is the OM boresight as measured in the OM reference frame
!*********************************************************************************************

    att_invbor = matmul(attmatrix, inv_boremat)
    omvec = matmul(att_invbor, ompoint)

    !************************************************************************
    ! converts the omvec cartesian vector to ra and dec spherical coordinates
    ! and check RA lies within sensible range
    ! ra_point and ra_dec refer to the pointing direction of the instrument
    !************************************************************************

    call sla_dcc2s(omvec, ra_point, dec_point)

    if (ra_point .lt. 0.0_double) ra_point = ra_point + 2.0_double * pi
   
  

    write(errstr, '(a, f8.4)') "RA of OM boresight: ", ra_point * radiansToDegrees
    call message(AppMsg, Verbose)
    write(errstr, '(a, f8.4)') "DEC of OM boresight: ", dec_point * radiansToDegrees
    call message(AppMsg, Verbose)


end subroutine OmPointing1


!****************************************************************
!
!
!***************************************************************
subroutine OmPointingDirection(dataSet, ra_point, dec_point)
    type(dataSetT), intent(in) :: dataSet

    real(double), intent(out) :: ra_point, dec_point  ! position of the OM tangent point
    real(double) :: rollangle                  ! the roll angle
   
    real(double) :: attmatrix(1 : 3, 1 : 3)    ! rotation matrix
    real(double) :: attmatrix1(0 : 2, 0 : 2)    ! rotation matrix

    real(double) :: att_invbor(3, 3)           !matrix for product of attitude*boresight
    real(double) :: ompoint(3)=(/1.0_double, 0.0_double, 0.0_double/)         !om boresight vector in OM frame
    real(double) :: pointvec(3)        ! s/c pointing
    real(double) :: pointvec2(3)       
    real(double), pointer :: boremat(:,:)       ! boresight offset matrix
    real(double) :: inv_boremat(3, 3)          ! matrix for inverse of boremat
    real(double) :: omvec(3)                   ! om pointing vector
  

    real(double) :: ra_scx, dec_scx, pa_scx  ! The ra, dec and pa of the spacecraft     
    real(double) :: r1                         ! modulus
    type (SpacecraftAttitudeType) attitude
     
    type(BoresightT) :: boresight    
    integer i, j    
    real(double) :: time

    ra_point = 0.0_double
    dec_point = 0.0_double


    !*******************************************************
    ! get the astrometry info keywords from the source list
    !*******************************************************
    ra_scx = real64Attribute(dataSet, "RA_SCX") * degreesToRadians   ! Spacecraft RA in radians
    dec_scx = real64Attribute(dataSet, "DEC_SCX") * degreesToRadians ! Spacecraft dec in radians
    pa_scx = real64Attribute(dataSet, "PA")        ! Spacecraft PA in degrees
 
 
    pa_scx = pa_scx * degreesToRadians
    rollAngle = pa_scx

    boresight = Cal_getBoresight(0.0_double);
    boremat => CAL_toDirectionCosineMatrix(boresight) 


    !***********************************************************
    ! Get the number of seconds since the mission reference time
    !***********************************************************
    time =  MissionTime(dataSet)

   
    
    !*******************************************************************
    ! need inverted boresight matrix which is obtained via the transpose. 
    !********************************************************************
    inv_boremat = TRANSPOSE(boremat)

    attitude = SpaceCraftAttitudeType(ra_scx, dec_scx, rollangle)

    !*******************************************************
    ! This routine requires a matrix of size 0:2, 0:2, but
    ! after that we want it to be of size 1:3, 1:3
    !******************************************************
    call OAL_toAttitudeMatrix(attitude, attmatrix1)
    
    forall( i = 1 : 3) 
        forall(j = 1 : 3) attmatrix(i, j) = attmatrix1(i - 1, j - 1)
    end forall

   
 
    !******************************************************************************
    ! converts RA and DEC from spherical cordinates to cartesian coordinates vector
    !*****************************************************************************
    call sla_dcs2c(ra_scx, dec_scx, pointvec)

    !********************************************************
    ! normalize the vector pointvec and return in pointvec2 -
    ! r1 holds the mdulus of the vector. pointvec2 will store
    ! the unit-vector in the direction of pointvec
    !********************************************************
    call sla_dvn(pointvec, pointvec2, r1)

    !
    !
    ! use SIMON'S BIT
    !
    !write(*,*)'using SIMONS BIT (essentially redundant now but keep for now
    !--------------------------------------------------------------------------------------------------------
    pointvec(1) = -pointvec2(1)
    pointvec(2) = -pointvec2(2)
    pointvec(3) =  pointvec2(3)
    !--------------------------------------------------------------------------------------------------------

 

    ! convert euler axis vector to a rotation matrix
     

    ! initialise output vector
    omvec = 0.0_double


!**********************************************************************************************
!  To get boresight correction, apply attitudematrix*(boresightmatrix)^(-1)*ompoint(=1,0,0)
!  Boresight * vector in SC frame gives vector in OM frame so boresight^(-1)*vector in OM frame gives 
!  vector in SC frame.  Then attitude * vector in S/C frame gives vector in Equatorial Reference Frame.
!
!  This creates omvec, the om boresight vector measured in the ERF
!  Note this process is applied here by first doing attmatrix*inv_boremat, then multiplying the 
!  ompoint vector by the product and decomposing to RA and DEC. The ompoint vector 
!  is the OM boresight as measured in the OM reference frame
!*********************************************************************************************

    att_invbor = matmul(attmatrix, inv_boremat)
    omvec = matmul(att_invbor, ompoint)

    !************************************************************************
    ! converts the omvec cartesian vector to ra and dec spherical coordinates
    ! and check RA lies within sensible range
    ! ra_point and ra_dec refer to the pointing direction of the instrument
    !************************************************************************

    call sla_dcc2s(omvec, ra_point, dec_point)

    if (ra_point .lt. 0.0_double) ra_point = ra_point + 2.0_double * pi
 

    write(errstr, '(a, f8.4)') "RA of OM boresight: ", (ra_point * radiansToDegrees)
    call message(AppMsg, Verbose)
    write(errstr, '(a, f8.4)') "DEC of OM boresight: ", (dec_point * radiansToDegrees)
    call message(AppMsg, Verbose)
   

end subroutine OmPointingDirection


!******************************************************************
!
!*****************************************************************
subroutine CorrectTheAstrometry(nrows,  rightAscension, declination, &
            posErr, magnitude, ra_point, dec_point, sourceListSet)     

    integer(int32), intent(in) :: nrows
    real(double), pointer ::  rightAscension(:), declination(:)
    real(single), pointer ::   posErr(:), magnitude(:)
    real(double), intent(in) :: ra_point, dec_point
    character(*), intent(in) :: sourceListSet
    logical :: status
    real(single) :: maxRADecErr = 1.0, maxRmsResidual = 1.5
    type(CorrectionData) :: corrData
    type(CatalogueData), pointer :: catData(:) => null()
    real(double), pointer ::  rightAscension1(:) => null(), declination1(:) => null()
    type(dataSetT) :: dataSetH
    type(TableT) :: sourceTable
    logical :: okay = .false.
    ! Keep the old ras and decs in case the fit is no good
    rightAscension1 => allocateDouble(1, size(rightAscension), "rightAscension1")
    declination1 => allocateDouble(1, size(rightAscension), "declination1")
    rightAscension1 = rightAscension * radiansToDegrees
    declination1 = declination * radiansToDegrees

    call CorrectAstrometry(nrows, rightAscension, declination, &
            posErr, magnitude, ra_point, dec_point,   sourceListSet, &
            status, corrData, catData)     
   
    
    if (status) then  ! The astrometry correction was successful
!        corrData%sigXShift =  sqrt(corrData%sigXShift *  corrData%sigXShift + 0.25)
!        corrData%sigYShift =  sqrt(corrData%sigYShift *  corrData%sigYShift + 0.25)

            maxRADEcErr = realParameter("maxradecerr")
            maxRmsResidual = realParameter("maxrmsres")

        ! Check if fitted parameters are okay 
        if (fitOkay(corrData, maxRADecErr, maxRmsResidual)) then
            rightAscension  = rightAscension * radiansToDegrees
            decLination = declination * radiansToDegrees
            okay = .true.

            call AddCorrectionColumns(sourceListSet, nrows, rightAscension, decLination, corrData)
        else
            call AddCorrectionColumns(sourceListSet, nrows, rightAscension1, decLination1, corrData)
   
        end if    

 
    else 
        call AddCorrectionColumns(sourceListSet, nrows, rightAscension1, decLination1, corrData)
   
        if (associated(catData)) then
        
        call pgbegin(0, "pgplot.ps/VC", 1, 1)


          
        call DrawGraph(nrows, real(rightAscension), &
                  real(decLination), "omatt", "RA", "DEC", 0.1, 0.9, 0.1, 0.9)
        call pgsci(3)
        call pgpoint(size(catData), real(catData%Ra * radiansToDegrees), &
                  real(catData%dec * radiansToDegrees), 3)
        call pgclos
       end if
       
   
    end if
    deallocate(rightAscension1, declination1)

    dataSetH = dataSet(sourceListSet, MODIFY)
    sourceTable = table(dataSetH, "SRCLIST")
   
    call setAttribute(sourceTable, "POSCOROK", okay, &
                              "ASTROMETRY CORRECTION OKAY")
    call setAttribute(sourceTable, "SYSERROR", 0.5, &
                              "Systematic catalogue position errors")

    call release(sourceTable)
    call release(dataSetH)
end subroutine CorrectTheAstrometry

!*******************************************
! Function ComputeEpoch
!******************************************
real(double) function  ComputeEpoch(dateObs)
   
    character(19), intent(in)  :: dateObs               
    real(double)  :: JulianDate0, JulianDate
    character(19)  :: missionReferenceTime = "1998-01-01T00:00:00"
    
    ! Compute the Julian date of the mission reference time.
    JulianDate0 = ModifiedJulianDate(missionReferenceTime)

    ! Compute the Julian date of the observation
    JulianDate = ModifiedJulianDate(dateObs)

    
    ComputeEpoch = 1998.0_double + (JulianDate - JulianDate0) / 365.25_double

end function ComputeEpoch


end module omatt_mod

